{{
NVVar Non-Volatile Variables V0.1

No purging of vars, no free-storage handling, no shrinking

*** THIS IS AN almost-BETA-VERSION! But it is getting beta er ... better

General:
--------
NVVar stores variables in EEPROM and can retrieve them back.
The distinguishing feature is, that it can handle different software-versions that sometimes add variables or change their size.
To do that, variables aren't stored in a block that is just an image of a contignous block in memory (HUB-RAM), but stores them with
tags.
A tag is an unique identifier the user assigns to a variable and that is also used to retrieve the data.
Additionally, you can assign a version number to the storage in EEPROM. This helps in detecting an older version of the NV-vars and
handling it in your code.

Usage:
------
1) Configure (Config) the "object"
2) Check wether the EEPROM is initialized (IsInit). If not, it has never been used to store data and you should *not* read your vars,
   but set their default
3) If the EEPROM is initialized, you can:
3a)  Retrieve the number of entries (NumberOfEntries)
3b)  Check if a tag exists (HasTag(<tag>))
3c)  Get the size of stored bytes for a tag (GetSize(<tag>) )
3d)  Read a variable from EEPROM by its tag (ReadVar(<tag>, <addressToWriteTo>, <sizeOfVariable>)
     Note, that the size of for destination var is truncated if the original size in EEPROM is smaller!
3e)  Write a variable (WriteVar(<tag>, <addressToReadFrom>, <size>) to EEPROM. If the var with the tag given doesn't exist yet, it is
     appened.
     If it does exist and the storage allocated in EEPROM differs from what you gave, the size is truncated.
4) Check the number of entries in EEPROM (Entries)
5) Set and get your version number (SetVersion(<versionNumber>), GetVersion)
6) Determine the size of storage required in EEPROM (GetFootprint)
7) To delete a NVVar-area and start from scratch just call Init. So Init is a *dangerous* function!
8) Do some cache configuration (SetCaching, GetCacheHint, SetCacheHint, ResetCache) See: Performance-tuning

 
Hint:
-----
You can save the same var with different tags
WriteVar(100, @theVar, 10)
is different to
WriteVar(101, @theVar, 10)
albeit the content *might* be the same and the address *is* the same. Addresses have not meaning except where to read from/write to.

Hint:
------
You can have more than one NVVar-area in EEPROM. It just has to be initialized in a different place.
Config(cNVVarEEPROMAddr, $8000, ...)
Config(cNVVarEEPROMAddr, $9000, ...)
But you can't have open more than one at a time. Ooops, might work with two or more NVVar-objects. Try at your own risk.

Performance-tuning:
-------------------
First of all, set the page size (cNVVarEEPROMPageSize) to the page-size of the EEPROM! If the page size is zero,
reads/writes will be done byte-wise and that is *very* slow. Page size depends on the chip you use. 64 bytes/page
might be a good guess.

There are two types of caching:
1.) A "tail-caching" that helps in finding the tail of the data in EEPROM when a new variable is added. It is always on, as
    there is no penalty for whatever pattern of updates / appends you use. There's only one delay when you do have a lot of existing entries
    and you do append something. This penalty is not a problem of the "tail-caching" but a problem of the single linked list.
    Maybe I'll add some tuning that will give an improvment if some reads have been done before
2.) Tag forward-caching that helps *tremendously* when reading vars in the sequence they were written. The gain gets lower, the further the
    searched entry is away from the last entry read.
    It gets a *real* pain if the vars are read in exactly the reverse order they were written to EEPROM.
    Just a number:
    With 1000 entries and 1000 reads, about 80MB have to be read just to find the 1000 entries
     
To solve the problem of 2), you have two possibilities:
1) Simply switch off forward-caching (SetCaching(true) or SetCaching(false)). That's the best methode for real random acces. Be aware
   that accesstime gets considerably high with many entries (100 or such) because the linked list has to be traversed from start on for
   every access (either read, write or append)
2) Use CacheHints (GetCacheHint(@<twoLongs>) or SetCacheHint(@<twoLongs>)). CacheHints set the forward-cache to some known (by you) location
   and starts searching from there (with or without forward-caching). If you do have local searches, it's best to get a cache-hint for the first entry,
   store it and call SetCacheHint and SetCaching(false).
   You can switch on forward-caching and use cache-hints, but the hint is forgotten after the first search and things
   get bad again if you do a backward search. 
   Do neverever modify the contens of cache-hint!
   If you are a geek, you can store cache-hints and retrieve them with a cache-hint. But that's going over my head!
3) Reset the forward-cache to the beginning of the list (ResetCache)
If you do have a lot of entries that are read many times, you do have to pay attention how the NVVar-storage is written the first time.
 
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Next version might break the current image in EEPROM !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

***************************************
Needing some more work:
* No CRC, data might be corruped without detection (CheckIntegrety etc).
* Unsave writing of size. Might completely corrupt the EEPROM block
* no re-alloc of unused space, no deletion of tags, no free-space housekeeping
***************************************

Just for your confusion, I internaly use "ID" what externally is called "tag". :-)


Layout of the block in EEPROM
  long Signature: to detect if the block is initialized
  long CRC: CRC over all entries BUGBUG: **implementation missing**
  long version: the user-defined version of the data
  long entries: number of entries in the block

  repeated entries times:
    long ID: the user-defined id of the variable
    long size: the size of the data
    n * byte data: datablock with the size size * bytes
}}

'Didn't test it recently, but if you want to use i2cObject instead of sdspifemto, look for these comments
'following an comment/uncomment them.

OBJ
'{sdspiFemto
  i2cObj : "sdspifemto.spin"
'}
{i2cObject
  i2cObj : "i2cObject.spin"
}
    
CON
  ''public constants. All beginning with "cNVVar...."
  cNVVarSignature = %00000001_00000011_00000111_00001111 'some funny "non-random" pattern. Chose your own if you don't like it
{i2cObject
  cNVVarEEPROMAddr = %1010_0000 'this is the standard EEPROM for propellers
  cNVVari2cSCL = 28
  cNVVari2cSDA = 29
  cAddrBits = 16
}
'{sdspiFemto
  cNVVarEEPROMAddr = i2cObj#bootAddr  
'}
  cNVVarEEPROMPageSize = 64 'set to zero if reads/writes have to be done byte-wise. Other value if a real block-read/write is wanted
                            'beware that blocksizes might change depending on the EEPROM? I don't know
                            'Set to the chip's size if there are no pages (RAM-like devices)
    
  cOffsCRC = 4
  cOffsVersion = 8
  cOffsEntries = 12
  cOffsData = 16
  cEntrySize = 4  '== long. size of a single entry (ID, size). All entries should have the same size
  cHeaderSizeL = 2 'size of header in longs
  cHeaderSizeB = 4 * cHeaderSizeL

                       
{i2cObject
  'assumes 80MHz
  'cNVVarReadTicks = 500    not waits after read required
  cWriteTicks = 70_000  'OK: 50_000
}
    
VAR
  long lBaseAddr      'with the sdspiFemto-driver, it includes the device-address too
  long lSignature
  byte lConfigured
' byte bNoCRC
  byte bCaching
  long lDevAddr
  long lNofEntriesC 'read-cache of NumberOfEntries
  long EndCache[2]  'chache of the farest end (near the tail).
                    'Exclusively handled by GetTail (albeit FindID could parially handle it too) [0]: index; [1]: offset
  long IDCache[2]   'make a vast improvment when IDs are retrieved in an increasing order. Otherwise a tad slower with
                    'pure random sequence
                    '[0]: index, [1]:offset
'{sdspiFemto
  long lCtrlBlock[2]
'}  
'-------------------------------------------------------------
PUB Config (devAddr, baseAddr, sig, WithCRC)

'{sdspiFemto
  lBaseAddr:= baseAddr + devAddr
'}
{i2cObject
  lBaseAddr:= baseAddr
}
  lSignature:= sig
' bNoCRC:= !WithCRC
  lDevAddr:= devAddr

{i2cObject
  i2cObj.Init(cNVVari2cSDA, cNVVari2cSCL, false)
  lConfigured:= i2cObj.devicePresent(lDevAddr)
}

'{sdspiFemto
  i2cObj.Start(@lCtrlBlock)
  lConfigured:= i2cObj.checkPresence(lDevAddr)
'}     
  EndCache[0]:= 1
  EndCache[1]:= cOffsData

  ResetCache
  bCaching:= true

  lNofEntriesC:= -1 'force read
    
  return lConfigured 
'-------------------------------------------------------------
PUB SetCaching(state)
  if state
    ResetCache    
  bCaching:= state
'-------------------------------------------------------------
PUB GetCacheHint(pCacheHint) ''must be pointer to two longs
  longmove(pCacheHint, @IDCache, 2)
'-------------------------------------------------------------
PUB SetCacheHint(pCacheHint) ''must be pointer to two longs
  longmove(@IDCache, pCacheHint, 2)
'-------------------------------------------------------------
PUB ResetCache
  IDCache[0]:= 0
  IDCache[1]:= cOffsData
'-------------------------------------------------------------
PUB Init(version) | zero   'sets the signature and does initial setup (call only once if IsInit returns false)
  if not lConfigured
    return FALSE

  zero:= 0
  SetEntries(0)
  WriteBlock(0, @lSignature, cEntrySize)
  WriteBlock(cOffsCRC, @zero, cEntrySize)  
  SetVersion(version)
    
  return true
'-------------------------------------------------------------
PUB NumberOfEntries
  if not lConfigured
    return FALSE

  if lNofEntriesC < 0
    ReadBlock(cOffsEntries, @lNofEntriesC, cEntrySize)
    
  return lNofEntriesC
'-------------------------------------------------------------
PUB IsInit : sig
  if not lConfigured
    return FALSE

  ReadBlock(0, @sig, cEntrySize)      
  return sig == lSignature
'-------------------------------------------------------------
PUB SetVersion(version)
  if not lConfigured
    return FALSE

  WriteBlock(cOffsVersion, @version, cEntrySize)
'-------------------------------------------------------------
PUB GetVersion : version
  if not lConfigured
    return FALSE

  ReadBlock(cOffsVersion, @version, cEntrySize)
  return version
'-------------------------------------------------------------
PUB GetSize(ID) : size | offset
  if not lConfigured
    return -1

  offset:= FindID(id)
  if offset < 0
    return -1

  ReadBlock(offset + cEntrySize, @size, cEntrySize)
  return size  
'-------------------------------------------------------------
PUB HasTag(id)
  if not lConfigured
    return FALSE

  return FindID(id) > 0
'-------------------------------------------------------------
PUB ReadVar(ID, pDest, size) | offset, actualSize  'read from EEPROM
  if not lConfigured
    return FALSE

  offset:= FindID(id)
  if offset < 0
    return -1

  ReadBlock(offset + cEntrySize, @actualSize, cEntrySize)
  if actualSize < size
    size:= actualSize 

  ReadBlock(offset + cHeaderSizeB, pDest, size)
   
  return size
'-------------------------------------------------------------
PUB WriteVar(id, pSrc, size) | offset, actualSize    'Write to EEPROM
  if not lConfigured
    return FALSE

  offset:= FindID(id)
  if offset < 0  'no such entry
    offset:= AddVar(ID, size)
    SetEntries(NumberOfEntries + 1)  'BUGBUG, not crash-proof this way round!

  'size adjust?
  'If for some reason (read: programmer's error) the desired size is bigger than the size of
  'the existing entry, we need to adjust it. This is just self-defense!
  ReadBlock(offset + cEntrySize, @actualSize, cEntrySize)
  if actualSize < size
    size:= actualSize
    
  WriteBlock(offset + cHeaderSizeB, pSrc, size) 
'-------------------------------------------------------------
PUB CheckIntegrity
  return true
'-------------------------------------------------------------
PUB GetFootprint
  if not lConfigured
    return FALSE

  return GetTail - cEntrySize  
'-------------------------------------------------------------
PRI GetTail : isAtOffset | numOfEntries, i, length, lastValid
'on exit points to where the ID has to be written

  numOfEntries:= NumberOfEntries

  if numOfEntries == 0
    return cOffsData 

  'handle the EndCache (not very elegant)
  if EndCache[0] > numOfEntries  'something went wrong!
    EndCache[0]:= 1
    EndCache[1]:= cOffsData
                                       
  isAtOffset:= EndCache[1]
  
  isAtOffset += cEntrySize 'point to length
  
  repeat i from EndCache[0] to numOfEntries
    'get the offset to the next entry ...
    lastValid:= isAtOffset    'BUGBUG could calculate outside of loop
    ReadBlock(isAtOffset, @length, cEntrySize) 
    isAtOffset += length  + cHeaderSizeB

  'let point EndCache to the last *valid* entry at *that* moment of time. At the next attempt, we will have
  ' to make two jumps along the list to reach the end                       
  EndCache[0]:= numOfEntries
  EndCache[1]:= lastValid - cEntrySize 
  
  return isAtOffset - cEntrySize 'pointed to size, but has to point to the ID 
'-------------------------------------------------------------
PRI AddVar(ID, size) : offset | header[cHeaderSizeL] 'be sure that the var doesn't exist! 
  'make a new entry at the end of the list and return the offset to the ID

  offset:= GetTail
  header[0]:= ID
  header[1]:= size
  WriteBlock(offset, @header, cHeaderSizeB)

  return offset  'point to ID
'-------------------------------------------------------------
PRI ReadBlock(offset, pDest, size) | i, addr

{i2cObject
  repeat i from 0 to size - 1
    byte[pDest + i]:= i2cObj.readLocation(lDevAddr, lBaseAddr + i + offset, cAddrBits)
}

'{sdspiFemto
  addr:= lBaseAddr + offset

  if cNVVarEEPROMPageSize > 0
    i2cObj.readEEPROM(addr, pDest, size)
  else ' real sequential read. It is -unlike the paged write- very simple
    repeat i from 0 to size - 1
      i2cObj.readEEPROM(addr++, pDest++, 1)
'}
'-------------------------------------------------------------
PRI WriteBlock(offset, pSrc, size) | i, addr, chunk

{i2cObject
  repeat i from 0 to size - 1
    i2cObj.writeLocation(lDevAddr, lBaseAddr + i + offset, byte[pSrc + i], cAddrBits)
    waitcnt(cWriteTicks + cnt)
}

'{sdspiFemto

  addr:= lBaseAddr + offset
  
  if cNVVarEEPROMPageSize > 0   'a real paged write
    repeat
      chunk:= cNVVarEEPROMPageSize - addr // cNVVarEEPROMPageSize
      if size =< chunk
        i2cObj.writeEEPROM(addr, pSrc, size)
        size:= 0
      else
        i2cObj.writeEEPROM(addr, pSrc, chunk)
        addr+= chunk
        pSrc+= chunk
        size-= chunk

      i2cObj.writeWait(lDevAddr)
    until size =< 0  'defensive ==
  else
    repeat i from 0 to size - 1
      i2cObj.writeEEPROM(addr++, pSrc++, 1)
      i2cObj.writeWait(lDevAddr)
'}
'-------------------------------------------------------------
PRI SetEntries(count)
  lNofEntriesC:= count
  WriteBlock(cOffsEntries, @lNofEntriesC, 4)
'-------------------------------------------------------------
PRI FindID(id) : atOffset | numOfEntries, i, index, header[cHeaderSizeL]

  numOfEntries:= NumberOfEntries

  if numOfEntries == 0
    return -1

  index:= IDCache[0]
  atOffset:= IDCache[1]         

  repeat i from 1 to numOfEntries
    ReadBlock(atOffset, @header, cHeaderSizeB)
    if header[0] == id
      'remember last hit when caching is on
      if bCaching
        IDCache[0]:= index
        IDCache[1]:= atOffset    
      return atOffset

    'set the offset to the next entry and try it there
    atOffset += header[1] + cHeaderSizeB

    if ++index => numOfEntries 'wrap-around
      index:= 0
      atOffset:= cOffsData

  return -1   
'-------------------------------------------------------------

   